import os
from pandas import read_csv, to_numeric
import numpy as np
from astropy.io import fits
import requests
import argparse

DEGTORAD = np.pi / 180.
RADtoASEC = (3600.0 * 180.0 / np.pi)
OFFSET = np.uint16(1000)
AMAX = np.uint16(60000)
R_RATIO = 10


def project_coords(ras, decs, ra, dec, s):
    delta_ra = (ras - ra) * DEGTORAD
    decs = decs * DEGTORAD
    dec = dec * DEGTORAD
    xx = np.cos(decs) * np.sin(delta_ra)
    yy = np.sin(dec) * np.sin(decs) + np.cos(dec) * \
        np.cos(decs) * np.cos(delta_ra)
    xi = (xx / yy) * RADtoASEC / s
    xx = np.cos(dec) * np.sin(decs) - np.sin(dec) * \
        np.cos(decs) * np.cos(delta_ra)
    eta = (xx / yy) * RADtoASEC / s
    return xi, eta


def gaussian2(x, y, x0, y0, s, A):
    return A * np.exp(-((x - x0)**2 + (y - y0)**2) / (2. * s**2))


def Run(ra=180, dec=-90, width=2000, V=15., height=1500, focal=1000., pixsize=5., FWHM=3., path=None):

    sampling = 3600. * 180. / np.pi * pixsize / focal / 1000.
    diag = (width ** 2 + height ** 2)**0.5
    r = diag * 0.5 * sampling / 60.

    query = 'http://vizier.u-strasbg.fr/viz-bin/asu-tsv?-source=NOMAD&-out.meta=-h-u-D&-out.add=_r&-sort=_r&-out=%20RAJ2000%20DEJ2000%20Vmag%20Bmag&-c='
    query += '{:10.8f}{:+10.8f}&-c.rm={:3.1f}&Vmag=<{:d}'.format(
        ra, dec, r, int(V))

    req = requests.request('GET', query)
    lines = [l for l in req.text.split('\n') if len(l) > 0]
    filename = 'cat_{:08.4f}_{:+07.4f}_r{:3.1f}_V{:3.1f}_s{:5.3f}.tsv'.format(
        ra, dec, r, V, sampling)
    if path is None:
        path = os.path.split(__file__)[0]

    filename = os.path.join(path, filename)
    with open(filename, 'w') as fin:
        for l in lines:
            fin.write('{:s}\n'.format(l))

    folder, filename = os.path.split(filename)
    os.chdir(folder)
    stars = read_csv(filename, delimiter='\t', comment='#',
                     names=['num', 'ras', 'decs', 'Vmag', 'Bmag'])
    stars = stars.iloc[1:]  # drop first lines with ---
    stars = stars.apply(to_numeric)

    print('Catalog fetched at ra:{:3.2f}, dec:{:+3.2f}, rad:{:3.1f}arcmin, mag:{:3.1f}, nbstars:{:d}'.format(
        ra, dec, r, V, len(stars)))

    stars['X'], stars['Y'] = project_coords(
        stars.ras, stars.decs, ra, dec, sampling)
    stars['X'] += float(width) * 0.5
    stars['Y'] += float(height) * 0.5

    img = OFFSET * np.ones([height, width], dtype=np.uint16)
    X, Y = np.meshgrid(np.linspace(0.5, width - 0.5, width),
                       np.linspace(0.5, height - 0.5, height))

    c = 0
    bright = np.min(stars.Vmag)
    faint = V + 1
    scale = AMAX / (-2.5 * np.log10(bright/faint))
    sigma = FWHM * .5 / (2. * np.log(2))**0.5 / sampling
    R = R_RATIO * sigma
    for _, row in stars.iterrows():
        x0 = row.X
        y0 = row.Y
        m = -2.5 * np.log10(row.Vmag / faint) * scale
        if x0 > 0 and x0 < width and y0 > 0 and y0 < height:
            mask = (X > x0 - R) & (X < x0 + R) & (Y > y0 - R) & (Y < y0 + R)
            X1 = X[mask]
            Y1 = Y[mask]
            img1 = gaussian2(X1, Y1, x0, y0, sigma, m).astype(np.uint16)
            img[mask] += img1
            c += 1
            print('{:5d}'.format(c), end='\r')
    print('total stars created: {:5d}'. format(c))
    a = 0
    hdu = fits.PrimaryHDU(np.flipud(img))
    hdr = hdu.header
    hdr['RA'] = (ra, '[deg] RA of telescope')
    hdr['DEC'] = (dec, '[deg] Declination of telescope')
    hdr['FOCALLEN'] = (focal, '[mm] Focal length')
    hdr['XPIXSZ'] = (pixsize, '[um] Pixel X axis size')
    hdr['YPIXSZ'] = (pixsize, '[um] Pixel Y axis size')
    hdr['SWCREATE'] = ('Starsynth v 0.0', 'Software that created this file')
    hdr['OBJECT'] = ('Synthetic field', '')
    hdr['DATE-OBS'] = ('2000-01-01T11:58:55.816', 'J2000 ref time')
    outname, _ = os.path.splitext(filename)
    outname += '.fit'
    hdu.writeto(outname, overwrite=True)
    # csvoutname, _ = os.path.splitext(filename)
    # csvoutname += '.csv'
    # stars.to_csv(csvoutname)


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='This script creates a synthetic star field and save the FITS file according to provided spec',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-r', '--ra', type=float,
                        help="Center RA in decimal deg", default=0.)
    parser.add_argument('-d', '--dec', type=float,
                        help="Center DEC in decimal deg", default=45.)
    parser.add_argument('-x', '--width', type=int,
                        help="Image width in pixels", default=2000)
    parser.add_argument('-y', '--height', type=int,
                        help="Image height in pixels", default=1500)
    parser.add_argument('-v', '--V', type=float,
                        help="Limit magnitude", default=15.)
    parser.add_argument('-F', '--focal', type=float,
                        help="Focal length in mm", default=1000.)
    parser.add_argument('-P', '--pixsize', type=float,
                        help="Pixel size in microns", default=5.)
    parser.add_argument('-f', '--FWHM', type=float,
                        help="Synthetic stars FWHM in arcsec", default=3.)
    parser.add_argument('-p', '--path', type=float,
                        help="output path", default=None)

    args = parser.parse_args()
    Run(**vars(args))
